/*  g_posix_shared.c - POSIX IPC SHARED MEMORY TEST
 *  
 *  (C) Copyright XII 2004 Michal 'GiM' Spadlinski <gim at skrzynka dot pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "g_posix_shared.h"
#include "dynstuff.h"
#include "g_ipc_wspolne.h"

extern const int g_data_format_s;
extern struct g_data_format parentp[2], childp[2];
int showtimes(char *str, list_t p);

#define G_NAME "/gimusia"

int mem_fd;
char *mem_addr;

int prefork_posix_shared(void)
{
	shm_unlink(G_NAME);
	if ((mem_fd = shm_open(G_NAME, O_RDWR | O_CREAT, 0600)) == -1)
	{
		perror("prefork shm_open");
		return -1;
	}
	ftruncate(mem_fd, 20 KB);
	mem_addr = mmap(0, 20 KB, PROT_READ|PROT_WRITE, MAP_SHARED, mem_fd, 0);
	shm_unlink(G_NAME);
	if (mem_addr == MAP_FAILED)
	{
		perror("prefork mmap failed");
		return -1;
	}

	memset(mem_addr, 0, 20 KB);

	return 0;
}

int child_posix_shared(int off)
{
	int i;
	long *lp = NULL;
	char *p, *q;
	struct timeval s, k;
	list_t sending = NULL, receiving = NULL;

	p = mem_addr;
	childp[off+0].type = 1;
	childp[off+0].id = getpid();
	q = childp[off+0].p;
	for (i=0; i<IPCQ_S; i++)
		*q++ = 'A';

	q = p + 10 KB;
	lp = (long *)p;
	DEB("[%d] child shared data prepared\n", getpid());
	for (i=0; i<MAX_QUERIES; i++)
	{
		TIMESTAMP(s);
		DEB("[%d] child sending data\n", getpid());
		memcpy(q + long_s, &(childp[off+0].id), IPCQ_S + long_s);
		memcpy(q, &(childp[off+0].type), long_s);
		DEB("[%d] child data sent %08lX\n", getpid(), *(unsigned long*)q);
		TIMESTAMP(k);

		DIFFTIME(k, s);
		list_add(&sending, &k, sizeof(struct timeval));

		TIMESTAMP(s);
		while (*lp != getpid()) sched_yield();
		memcpy (&(childp[off+1]), lp, IPCQ_S + 2 * long_s);
		*lp = 0;

		TIMESTAMP(k);
		DIFFTIME(k, s);
		list_add(&receiving, &k, sizeof(struct timeval));

		childp[off+1].p[IPCQ_S - 1] = 0; 
		DEB("[%d] child received data from %li\n", getpid(), childp[off+1].id);
	}
	showtimes("[+] posx sh memory   sending", sending);
	showtimes("[+] posx sh memory receiving", receiving);

	list_destroy(sending, 1);
	list_destroy(receiving, 1);

	return 0;
}

int parent_posix_shared(int pid)
{
	char *p;
	long *lp;

	p = mem_addr;
	lp = (long *)(mem_addr + 10 KB);
	while (1)
	{
		DEB("[%d] parent shared waiting for data %ld\n", 
				getpid(), *lp);
		if (*lp == 1)
		{
			/* receive */
			memcpy (&(parentp[0]), lp, IPCQ_S + 2 * long_s);
			*lp = 0;
			DEB("[%d] parend shared data received from %li",
					getpid(), parentp[0].id);

			/* send */
			parentp[0].type = parentp[0].id;
			parentp[0].id = getpid();
			memcpy (mem_addr + long_s, &(parentp[0].id), IPCQ_S + long_s);
			memcpy (mem_addr, &(parentp[0].type), long_s);
		} else {
			if (waitpid(pid, NULL, WNOHANG) != 0)
				break;
			sched_yield();
		}
	}

	DEB("[%d] parent child exited\n", getpid());
	DEB("[%d] parent detached\n", getpid());

	return 0;
}
